"""Python script that produces reStructuredText out of JUnit/xUnit output (XML)"""
import argparse
import logging
import xml.etree.ElementTree as ET
from collections import namedtuple
from pathlib import Path

from mako.exceptions import RichTraceback
from mako.runtime import Context
from mako.template import Template
from pkg_resources import DistributionNotFound, require

TraceableInfo = namedtuple("TraceableInfo", ['matrix_prefix', 'unit_or_integration', 'header_prefix'])
UTEST = TraceableInfo('UTEST_', 'unit', '_unit_test_report_')
ITEST = TraceableInfo('ITEST_', 'integration', '_integration_test_report_')
TEMPLATE_FILE = Path(__file__).parent.joinpath('xunit2rst.mako')


def render_template(destination, **kwargs):
    """ Renders the Mako template, and writes output file to the specified destination.

    Args:
        destination (Path): Location of the output file.
        **kwargs (dict): Variables to be used in the Mako template.

    Raises:
        ERROR: Error log containing information about the line where the exception occurred.
        Exception: Re-raised Exception coming from Mako template.
    """
    destination.parent.mkdir(parents=True, exist_ok=True)
    with open(str(destination), 'w', newline='\n', encoding='utf-8') as rst_file:
        template = Template(filename=str(TEMPLATE_FILE), output_encoding='utf-8', input_encoding='utf-8')
        try:
            template.render_context(Context(rst_file, **kwargs))
        except Exception as exc:
            traceback = RichTraceback()
            logging.error("Exception raised in Mako template, which will be re-raised after logging line info:")
            logging.error("File %s, line %s, in %s: %r", *traceback.traceback[-1])
            raise exc


def generate_xunit_to_rst(input_file, rst_file, itemize_suites, failure_message, *prefix_args):
    """ Processes input arguments, calls mako template function while passing all needed parameters.

    Args:
        input_file (Path): Path to the input file (.xml).
        rst_file (Path): Path to the output file (.rst).
        itemize_suites (bool): True for itemization of testsuite elements, False for testcase elements.
        failure_message (bool): True if failure messages are to be included in the item's body, False otherwise.
    """
    test_suites, prefix_set = parse_xunit_root(input_file)

    prefix_set, prefix = build_prefix_and_set(test_suites, prefix_set, *prefix_args)

    report_name = rst_file.stem
    if report_name.endswith('_report'):
        report_name = report_name[:-len('_report')]

    render_template(
        rst_file,
        test_suites=test_suites,
        report_name=report_name,
        info=prefix_set,
        prefix=prefix,
        itemize_suites=itemize_suites,
        failure_message=failure_message,
    )


def parse_xunit_root(input_file):
    """
    This function parses the root element of the XML file and returns a testsuites root element and the set of prefixes
    to use.

    Args:
        input_file (Path): Path to the input file (.xml).

    Returns:
        xml.etree.ElementTree.Element: Root element of the element tree with 'testsuites' as tag.
        TraceableInfo: Namedtuple holding the prefixes to use for building traceability output.
    """
    tree = ET.parse(str(input_file))
    root_input = tree.getroot()
    if root_input.tag != 'testsuites':
        test_suites = ET.Element("testsuites")
        test_suites.append(root_input)
        prefix_set = ITEST
    else:
        test_suites = root_input
        prefix_set = UTEST

    return test_suites, prefix_set


def build_prefix_and_set(test_suites, prefix_set, prefix, trim_suffix, unit_or_integration):
    """ Builds the prefix and prefix_set variables based on the input parameters.

    Args:
        test_suites (xml.etree.ElementTree.Element): Root element of the element tree with 'testsuites' as tag.
        prefix_set (TraceableInfo): Namedtuple holding the default prefix to use for building traceability output.
        prefix (str): Prefix to add to item IDs. In case of an empty string, the prefix from the element's name will be
            used, or the default prefix otherwise.
        trim_suffix (bool): Whether to trim the suffix of the prefix or not.
        unit_or_integration (None/str): None if the script's discernment shall be used, otherwise a string starting
            with 'u' or 'i', indicating unit test report or integration test report respectively as input.

    Returns:
        prefix_set (TraceableInfo): Namedtuple holding the prefixes to use for building traceability output.
        prefix (str): Prefix to add to item IDs.
    """
    if prefix.endswith('_-') and trim_suffix:
        prefix = prefix.rstrip('_-') + '-'

    base_prefix_on_set = False
    if not prefix:
        item_name_halves = list(test_suites)[-1].attrib['name'].split('.')[-1].split('-')
        if len(item_name_halves) > 1:
            prefix = item_name_halves[0] + '-'
        else:  # no prefix in name
            prefix = prefix_set.matrix_prefix
            base_prefix_on_set = True

    prefix_set = verify_prefix_set(prefix_set, prefix, unit_or_integration)
    if base_prefix_on_set:
        prefix = prefix_set.matrix_prefix
    prefix = prefix.rstrip('_')
    prefix = prefix + '-' if not prefix.endswith('-') else prefix
    return prefix_set, prefix


def verify_prefix_set(prefix_set, prefix, unit_or_integration):
    """
    The unit-or-integration test input argument has the highest priority, followed by the first letter in the prefix,
    and lastly the script will interpret a test report as a unit test report if it contains a 'testsuites' element,
    integration test report otherwise.

    Args:
        prefix_set (TraceableInfo): TraceableInfo UTEST or ITEST decided by the presence or lack of a root element
            'testsuites'.
        prefix (str): Prefix that will be used in the Mako template.
        unit_or_integration (None/str): None if the script's discernment shall be used, otherwise a string starting
            with 'u' or 'i', indicating unit test report or integration test report respectively as input.

    Raises:
        ValueError: The unit-or-integration argument is used, but is invalid.
    """
    if isinstance(unit_or_integration, str):
        discerning_letter = '' if not unit_or_integration else unit_or_integration.lower()[0]
    else:
        discerning_letter = prefix.lower()[0]
    if discerning_letter in 'iu':
        prefix_set = ITEST if discerning_letter == 'i' else UTEST
    elif unit_or_integration is not None:
        raise ValueError("Value for --unit-or-integration input argument is invalid: expected 'u' or 'i'; got {!r}."
                         .format(unit_or_integration))
    return prefix_set


def create_parser():
    """ Creates and returns the ArgumentParser instance to be used """
    try:
        version = require('mlx.xunit2rst')[0].version
    except DistributionNotFound:
        version = '0.0.0.dev'
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-i', '--input',
                            action='store',
                            required=True,
                            dest='input_file',
                            type=Path,
                            help='The input XML file',)
    arg_parser.add_argument('-o', '--output',
                            action='store',
                            required=True,
                            dest='rst_output_file',
                            type=Path,
                            help='The output RST file',)
    arg_parser.add_argument('-s', '--itemize-suites',
                            action='store_true',
                            help="Flag to itemize testsuite elements instead of testcase elements.")
    arg_parser.add_argument('-p', '--prefix',
                            action='store',
                            default="",
                            help='Optional prefix to add to item IDs',)
    arg_parser.add_argument("--trim-suffix", action='store_true',
                            help="If the suffix of the --prefix argument ends with '_-' it gets trimmed to '-'")
    arg_parser.add_argument("--unit-or-integration", action='store',
                            help="Optional: give value starting with 'u' or 'i' if the the script's discernment is "
                                 "wrong.")
    arg_parser.add_argument("-f", "--failure-message", action="store_true",
                            help="Include the error message in case of test failure in the item's body.")
    arg_parser.add_argument('-v', '--version',
                            action='version',
                            version='%(prog)s {}'.format(version),)
    return arg_parser


def main():
    """Main function"""
    arg_parser = create_parser()
    args = arg_parser.parse_args()

    generate_xunit_to_rst(
        args.input_file,
        args.rst_output_file,
        args.itemize_suites,
        args.failure_message,
        args.prefix,
        args.trim_suffix,
        args.unit_or_integration,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
''' Test suite for error handling for Mako template '''
from unittest import TestCase
from pathlib import Path
from nose.tools import assert_raises
from mlx.xunit2rst import render_template, ITEST

TEST_OUT_DIR = Path(__file__).parent / 'test_out'


def test_mako_error_handling():
    ''' Tests error logging and re-raising of an Exception in Mako template by intentionally not passing a variable '''
    kwargs = {
        'report_name': 'my_report',
        'info': ITEST,
        'prefix': 'MAKO_TEST-',
    }
    test_case = TestCase()
    with test_case.assertLogs() as log_cm:
        with assert_raises(TypeError):
            render_template((TEST_OUT_DIR / 'never_created_file.rst'), **kwargs)
    test_case.assertIn('Exception raised in Mako template, which will be re-raised after logging line info:',
                       log_cm.output[0])
    test_case.assertIn('File ', log_cm.output[-1])
    test_case.assertIn('line ', log_cm.output[-1])
    test_case.assertIn("in render_body: '% for suite in test_suites:'", log_cm.output[-1])



#!/usr/bin/env python3
''' Acceptance test suite for mlx.xunit2rst

The tests in this file are pure black box tests. They just check whether the
outputs of the tool match the reference output files exactly.
'''
import filecmp
from pathlib import Path

import nose
from nose.tools import with_setup

from mlx.xunit2rst import create_parser, generate_xunit_to_rst

TOP_DIR = Path(__file__).parents[1]
TEST_OUT_DIR = Path(__file__).parent / 'test_out'
TEST_IN_DIR = Path(__file__).parent / 'test_in'


def setup():
    ''' Setup function

    This function creates the output directory for mlx.xunit2rst to put the
    output files
    '''
    TEST_OUT_DIR.mkdir(exist_ok=True)


def xunit2rst_check(input_xml, output_rst, itemize_suites=False, prefix='', trim=False, u_or_i=None, failures=False):
    ''' Helper function for testing whether mlx.xunit2rst produces the expected output '''
    arg_parser = create_parser()
    command = ['-i', input_xml, '-o', output_rst]
    if itemize_suites:
        command.append('--itemize-suites')
    if prefix:
        command.extend(['-p', prefix])
    if trim:
        command.append('--trim-suffix')
    if u_or_i is not None:
        command.extend(['--unit-or-integration', u_or_i])
    if failures:
        command.extend(['-f'])
    print(command)
    args = arg_parser.parse_args(command)

    generate_xunit_to_rst(
        args.input_file,
        args.rst_output_file,
        args.itemize_suites,
        args.failure_message,
        args.prefix,
        args.trim_suffix,
        args.unit_or_integration,
    )


@with_setup(setup)
def test_xunit_default():
    '''Tests based on reports generated by `robot --xunit` '''
    file_name = 'itest_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='ITEST-')

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_custom_prefix():
    '''Tests based on reports generated by `robot --xunit` '''
    file_name = 'itest_lin_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='ITEST_LIN-')

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_no_prefix():
    '''Tests based on reports generated by `robot --xunit` '''
    file_name = 'itest_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_trim_suffix():
    '''Tests based on reports generated by `robot --xunit` '''
    file_name = 'itest_lin_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='ITEST_LIN_-', trim=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_trim_suffix_not_needed():
    '''Tests based on reports generated by `robot --xunit` '''
    file_name = 'itest_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='ITEST-', trim=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_overwrite_discernment():
    '''Tests based on reports generated by `robot --xunit` - forcing script to treat input as unit test report '''
    rst_file_name = '{}.rst'.format('itest_as_utest_report')
    xml_file_name = '{}.xml'.format('itest_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='ITEST-', u_or_i='u')

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_junit_default():
    '''Tests based on utest reports in JUnit format - itemizing testcase elements '''
    file_name = 'utest_my_lib_report'
    rst_file_name = '{}.rst'.format(file_name)
    xml_file_name = '{}.xml'.format(file_name)
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, itemize_suites=False, prefix='')

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_junit_itemize_testsuites():
    '''Tests based on utest reports in JUnit format - itemizing testsuite elements'''
    rst_file_name = '{}.rst'.format('utest_my_lib_suites_report')
    xml_file_name = '{}.xml'.format('utest_my_lib_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, itemize_suites=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_junit_prefix():
    '''Tests based on utest reports in JUnit format - adding prefix via input arg'''
    rst_file_name = '{}.rst'.format('utest_my_lib_suites_report')
    xml_file_name = '{}.xml'.format('utest_my_lib_no_prefix_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='UTEST_MY_LIB-', itemize_suites=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_junit_override_xml_prefix():
    '''Tests based on utest reports in JUnit format - adding prefix via input arg'''
    rst_file_name = '{}.rst'.format('utest_override_prefix_report')
    xml_file_name = '{}.xml'.format('utest_my_lib_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, prefix='OVERRIDING-')

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_xunit_failure_messages():
    '''Tests based on itest reports in xUnit format - including failure messages'''
    rst_file_name = '{}.rst'.format('itest_lin_report_failures')
    xml_file_name = '{}.xml'.format('itest_lin_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, failures=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


@with_setup(setup)
def test_junit_failure_messages():
    '''Tests based on utest reports in JUnit format - including failure messages'''
    rst_file_name = '{}.rst'.format('utest_my_lib_report_failures')
    xml_file_name = '{}.xml'.format('utest_my_lib_no_prefix_report')
    input_xml = str(TEST_IN_DIR / xml_file_name)
    output_rst = str(TEST_OUT_DIR / rst_file_name)
    xunit2rst_check(input_xml, output_rst, failures=True, itemize_suites=True)

    reference_rst = str(TEST_IN_DIR / rst_file_name)
    assert filecmp.cmp(output_rst, reference_rst)


if __name__ == '__main__':
    nose.main()



#!/usr/bin/env python3
''' Test suite for functions that set the prefix and prefix_set variables '''
from pathlib import Path

import nose
from nose.tools import assert_equal, assert_not_equal, assert_raises

from mlx.xunit2rst import build_prefix_and_set, parse_xunit_root, verify_prefix_set, ITEST, UTEST

TEST_IN_DIR = Path(__file__).parent / 'test_in'


def test_build_prefix_and_set_utest_default():
    ''' Use default prefix for unit test reports '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'utest_my_lib_no_prefix_report.xml')
    assert_equal(initial_prefix_set, UTEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, '', True, None)
    assert_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix, 'UTEST-')


def test_build_prefix_and_set_itest_default():
    ''' Use default prefix for integration test reports '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'itest_report.xml')
    assert_equal(initial_prefix_set, ITEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, '', True, None)
    assert_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix, 'ITEST-')


def test_build_prefix_and_set_from_name():
    ''' Get prefix from element name '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'utest_my_lib_report.xml')
    assert_equal(initial_prefix_set, UTEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, '', True, None)
    assert_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix, 'UTEST_MY_LIB-')


def test_build_prefix_and_set_from_arg():
    ''' Get prefix from input argument `--prefix` and trim suffix of prefix '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'utest_my_lib_report.xml')
    assert_equal(initial_prefix_set, UTEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, 'TEST_MY_LIB_-', True, None)
    assert_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix, 'TEST_MY_LIB-')


def test_build_prefix_and_set_from_arg_swap_set():
    '''
    Get prefix from input argument `--prefix` and base prefix_set on its first letter.
    Don't trim suffix of prefix.
    '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'itest_report.xml')
    assert_equal(initial_prefix_set, ITEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, 'UTEST_MY_LIB_-', False, None)
    assert_not_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix_set, UTEST)
    assert_equal(prefix, 'UTEST_MY_LIB_-')


def test_build_prefix_and_set_priority():
    ''' Argument unit_or_integration must have the highest priority for determining the correct prefix_set. '''
    test_suites, initial_prefix_set = parse_xunit_root(TEST_IN_DIR / 'utest_my_lib_report.xml')
    assert_equal(initial_prefix_set, UTEST)

    prefix_set, prefix = build_prefix_and_set(test_suites, initial_prefix_set, 'UTEST_HOWDY-', False, 'i')
    assert_not_equal(prefix_set, initial_prefix_set)
    assert_equal(prefix_set, ITEST)
    assert_equal(prefix, 'UTEST_HOWDY-')


def test_verify_prefix_set_u_or_i():
    '''
    Tests verify_prefix_set function. The unit_or_integration argument should have the highest priority and must
    start with u or i (case-insensitive). The prefix argument has the second highest priority. A last resort is to
    keep the input prefix_set.
    '''
    assert_equal(verify_prefix_set(UTEST, 'UTEST-', 'Itest'), ITEST)
    assert_equal(verify_prefix_set(UTEST, 'UTEST-', 'i'), ITEST)
    assert_equal(verify_prefix_set(UTEST, '', 'i'), ITEST)
    assert_equal(verify_prefix_set(ITEST, 'ITEST-', 'Utest'), UTEST)
    assert_equal(verify_prefix_set(UTEST, 'ITEST-', 'u'), UTEST)
    assert_equal(verify_prefix_set(UTEST, 'UTEST-', 'u'), UTEST)
    assert_equal(verify_prefix_set(UTEST, 'BLAH-', None), UTEST)
    assert_equal(verify_prefix_set(ITEST, 'BLAH-', None), ITEST)
    assert_equal(verify_prefix_set(UTEST, 'ITEST-', None), ITEST)
    assert_equal(verify_prefix_set(ITEST, 'UTEST-', None), UTEST)
    with assert_raises(ValueError):
        verify_prefix_set(UTEST, 'UTEST-', 't')
    with assert_raises(ValueError):
        verify_prefix_set(ITEST, 'ITEST', 't')
    with assert_raises(ValueError):
        verify_prefix_set(ITEST, '', 't')


if __name__ == '__main__':
    nose.main()
